/*
 * observer.h
 *
 *  Created on: Apr 12, 2011
 *      Author: lipari
 */

#ifndef OBSERVER_H_
#define OBSERVER_H_

#include <map>
#include <vector>

//#include <iostream>
//using namespace std;

namespace OOSD {

	class ObserverConnectorInterface {
		bool todelete;
	public:
		ObserverConnectorInterface() : todelete(false) {}
		virtual void update() = 0;
		virtual void *getPointer() = 0;
		bool toDelete() { return todelete; }
		virtual void setDeletion() { todelete = true; }
		virtual ~ObserverConnectorInterface() {}
	};

	/**
	 * A ObserverConnector is a function object whose only job is to
	   "connect" a subject of a certain type (template parameter Subject)
	   to an observer (template parameter Observer).  The user only needs
	   to call the constructor to create a ObserverConnector.
   */
	template <class Subject, class Observer>
	class ObserverConnector : public ObserverConnectorInterface {
		Subject &sub;
		Observer &obs;
	public:
		ObserverConnector(Subject &s, Observer &o) : sub(s), obs(o) {
			s.addObserver(*this);
		}
		void update() {
			obs.update(sub);
		}
		void *getPointer() { return &obs; }
	};


	class BaseObservable {
		typedef std::map<void*, ObserverConnectorInterface *> ObserverList;
		typedef std::map<void*, ObserverConnectorInterface *>::const_iterator const_iterator;
		typedef std::map<void*, ObserverConnectorInterface *>::iterator iterator;
		typedef std::vector<ObserverConnectorInterface *> DeferredList;
		typedef std::vector<ObserverConnectorInterface *>::iterator deferred_iterator;

		ObserverList observers;
		DeferredList toAttach;
		bool notification;
		struct BooleanGuard {
			bool &flag;
			BooleanGuard(bool &f) : flag(f) { flag = true; }
			~BooleanGuard() { flag = false; }
		};
	protected:
		void deferredDetach() {
			for (iterator i=observers.begin(); i!=observers.end();)
				if (i->second->toDelete()) {
					delete i->second;
					iterator tmp = i++;
					observers.erase(tmp);
				}
				else ++i;
		}

		void deferredAttach() {
			for (deferred_iterator i=toAttach.begin(); i!=toAttach.end(); ++i)
				observers[(*i)->getPointer()] = *i;
			toAttach.clear();
		}

		void notifyAll() {
			BooleanGuard guard(notification);

			for (iterator i=observers.begin(); i!=observers.end(); ++i)
				if (!i->second->toDelete()) i->second->update();

			deferredDetach();
			deferredAttach();
		}


	public:
		BaseObservable() : notification(false) {}

		void addObserver(ObserverConnectorInterface &o) {
			if (notification) toAttach.push_back(&o);
			else observers[o.getPointer()] = &o;
		}

		void detachObserver(void *address) {
			iterator i = observers.find(address);
			if (i != observers.end()) {
				if (notification) i->second->setDeletion();
				else {
					delete i->second;
					observers.erase(i);
				}
			}
		}

		virtual ~BaseObservable() {
			ObserverList::const_iterator i;
			for (i=observers.begin(); i!=observers.end(); ++i)
				delete i->second;
		}
	};

	template <class Subject, class Observer>
	void register_observer(Subject &s, Observer &o) {
		new ObserverConnector<Subject, Observer>(s, o);
	}

	template <class Subject, class Observer>
	void unregister_observer(Subject &s, Observer &o) {
		s.detachObserver(&o);
	}
}


#endif /* OBSERVER_H_ */
